/**
 * Copyright (c) 2007 University of Patras Computer Engineering and Informatics Department Greece
 *
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
  */

#include <stdlib.h>
#include "config.h"
#include "agent.h"
#include "random.h"
#include "rtp.h"


int hdr_rtp::offset_;
/** the RTP header class */
class RTPHeaderClass : public PacketHeaderClass {
public: 
        RTPHeaderClass() : PacketHeaderClass("PacketHeader/RTP",
					     sizeof(hdr_rtp)) {
		bind_offset(&hdr_rtp::offset_);
	}
} class_rtphdr;
/** creates an RTPAgent
    returns the newly created Agent
*/ 
static class RTPAgentClass : public TclClass {
public:
	RTPAgentClass() : TclClass("Agent/RTP") {}
	TclObject* create(int, const char*const*) {
		return (new RTPAgent());

	}
} class_rtp_agent;

/** constructor 
   initializes the RTPAgent intance
*/
RTPAgent::RTPAgent() : Agent(PT_RTP), session_(0), lastpkttime_(-1e6),
    running_(0),  rtp_timer_(this)
{
	bind("seqno_", &seqno_);
	bind_time("interval_", &interval_);
	bind("packetSize_", &size_);
	bind("maxpkts_", &maxpkts_);
	bind("random_", &random_);
	timestamp_ = 0;
	
}
/** starts the RTP Agent*/
void RTPAgent::start()
{
        running_ = 1;
        sendpkt();
        rtp_timer_.resched(interval_);
}

/** stops the RTP Agent*/
void RTPAgent::stop()
{
//	printf("inside stop localsrc_ %d\n", session_->localsrc_->srcid());
        rtp_timer_.force_cancel();
        finish();
}

/** sends the packets */
void RTPAgent::sendmsg(int nbytes, const char* )
{
        Packet *p;
        int n;

	assert (size_ > 0);

        if (++seqno_ < maxpkts_) {
		n = nbytes / size_;
	
                if (nbytes == -1) {
                        start();
                        return;
                }
                while (n-- > 0) {
                        p = allocpkt();
                        hdr_rtp* rh = hdr_rtp::access(p);
                        rh->seqno() = seqno_;
                        target_->recv(p);
                }
                n = nbytes % size_;
                if (n > 0) {
                        p = allocpkt();
                        hdr_rtp* rh = hdr_rtp::access(p);
                        rh->seqno() = seqno_;
                        target_->recv(p);
                }
                idle();
        } else {
                finish();
                // xxx: should we deschedule the timer here? */
        };
}

/** when the RTPTimer times out the RTP Agent sends an RTP packet
   when there is a session (multicast session) the sending source is
   apdating the number of RTP packets that sends
*/
void RTPAgent::timeout(int)
{
	if (running_) {
		sendpkt();
		if (session_)
			session_->localsrc_update(size_);
			session_->localsrc_update_nbytes(size_);
			
		double t = interval_;
		if (random_)
			/* add some zero-mean white noise */
			t += interval_ * Random::uniform(-0.5, 0.5);
		rtp_timer_.resched(t);
	}
}

/**
 * finish() is called when we must stop (either by request or because
 * we're out of packets to send.
 */
void RTPAgent::finish()
{
        running_ = 0;
        Tcl::instance().evalf("%s done", this->name());
}

/**
 * 
 * @param delta 
 */
void RTPAgent::advanceby(int delta)
{
        maxpkts_ += delta;
        if (seqno_ < maxpkts_ && !running_)
                start();
}               

/** the receive method for RTP packets. It simples passes
   the packet to the instance of the RTPSession class. Ottherwise
   the mothod does nothing
*/
void RTPAgent::recv(Packet* p, Handler*)
{
	if (session_)
		session_->recv(p, 0);
	else
		Packet::free(p);
}
/* to handle events from the TCL classes or the TCL scripts */
int RTPAgent::command(int argc, const char*const* argv)
{
	if (argc == 2) {
		if (strcmp(argv[1], "rate-change") == 0) {
			rate_change();
			return (TCL_OK);
		} else if (strcmp(argv[1], "start") == 0) {
                        start();
                        return (TCL_OK);
                } else if (strcmp(argv[1], "stop") == 0) {
                        stop();
                        return (TCL_OK);
                }
	} else if (argc == 3) {
		if (strcmp(argv[1], "session") == 0) {
			session_ = (RTPSession*)TclObject::lookup(argv[2]);
			return (TCL_OK);
		} else if (strcmp(argv[1], "advance") == 0) {
                        int newseq = atoi(argv[2]);
                        advanceby(newseq - seqno_);
                        return (TCL_OK); 
                } else if (strcmp(argv[1], "advanceby") == 0) {
                        advanceby(atoi(argv[2]));
                        return (TCL_OK);
                }       
	}
	return (Agent::command(argc, argv));
}

/** 
 * We modify the rate in this way to get a faster reaction to the a rate
 * change since a rate change from a very low rate to a very fast rate may 
 * take an undesireably long time if we have to wait for timeout at the old
 * rate before we can send at the new (faster) rate.
 */
void RTPAgent::rate_change()
{
	rtp_timer_.force_cancel();
	
	double t = lastpkttime_ + interval_;
	
	double now = Scheduler::instance().clock();
	if ( t > now)
		rtp_timer_.resched(t - now);
	else {
		sendpkt();
		rtp_timer_.resched(interval_);
	}
}
/** allocates the RTP packet that has to be sent */
void RTPAgent::sendpkt()
{
	Packet* p = allocpkt();
	lastpkttime_ = Scheduler::instance().clock();
	timestamp_ = lastpkttime_;
	makepkt(p);
	target_->recv(p, (Handler*)0);
}
/** creates the RTP packet and  fills in the packet header fields */
void RTPAgent::makepkt(Packet* p)
{
	
	hdr_rtp *rh = hdr_rtp::access(p);
	/* Fill in srcid_ and seqno */
	rh->seqno() = seqno_++;
	rh->srcid() = session_ ? session_->srcid() : 0;
	rh->timestamp()= timestamp_;
	
	
}

void RTPTimer::expire(Event* /*e*/) {
        a_->timeout(0);
}


