/**
 * Copyright (c) 2007 University of Patras Computer Engineering and Informatics Department Greece
 *
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef ns_rtp_h
#define ns_rtp_h

#include "config.h"
#include "object.h"
#include "agent.h"
#include "timer-handler.h"
#include <list>	

#define RTP_M 0x0080 // marker for significant events
class RTPReceiver;
/** rtcpup sender reports */
struct sender_report {
	/** the id of the originator of this SR */
        u_int32_t sender_srcid_;
	/** total RTP packets sent since the beginning of the session */
	int pkts_sent_;
	/** total octets sent since the beginning of the session */
        int octets_sent_;
	/** the receivers the originator of this report serves */
	RTPReceiver* rcvrs_;
	/* per-field member functions */
	u_int32_t& sender_srcid() { return (sender_srcid_); }
	/** return the total RTP packets sent since the beginning of the session */
	int& pkts_sent() { return (pkts_sent_); }
	/** return the total octets sent since the beginning of the session */
	int& octets_sent() { return (octets_sent_); }
}; 

/** rtcpup receiver reports */
struct receiver_report {
	/** total number of RTP packets lost since the beginning of the session */
	int cum_pkts_lost_;
	/** the TCP friendly bandwidth share */
        double R_tcp_;
	/** last Sender Report time */
        double LSR_;
	/** delay since last sender report */
        double DLSR_;
	/** the delay jitter after filtering */
	double jitter_;
	int  bye_;
	/* per-field member functions */
	/** return the total number of RTP packets lost since the beginning of the session */
	int& cum_pkts_lost() { return (cum_pkts_lost_); }
        /** return the the TCP friendly bandwidth share */
	double& R_tcp() { return (R_tcp_); }
	/** return the last Sender Report time */
	double& LSR() { return (LSR_); }
	/** return delay since last sender report */
	double& DLSR() { return (DLSR_); }
	/** return the the delay jitter after filtering */
	double& jitter() { return (jitter_); }
        int& bye() { return (bye_); }

}; 

/** rtp packet header */
struct hdr_rtp {
	/** the unique source id for this session */
	u_int32_t srcid_;
	/** sequence number of RTP and RTCP packets */
	int seqno_;
	/**rtp flags indicating significant event(begining of talkspurt) */
	u_int16_t flags_; 
	/** timestamp of the RTP packet */
	double timestamp_;
	static int offset_;
	inline static int& offset() { return offset_; }
	inline static hdr_rtp* access(const Packet* p) {
		return (hdr_rtp*) p->access(offset_);
	}
	/* per-field member functions */
	/** return the the unique source id for this session */
	u_int32_t& srcid() { return (srcid_); }
	/** return the sequence number of RTP and RTCP packets */
	int& seqno() { return (seqno_); }
	u_int16_t& flags() { return (flags_); }
	/** return the timestamp of the RTP packet */
	double& timestamp() { return (timestamp_); }
	/** the rtcp sender report */
	sender_report* sr_;
	/** the rtcp receiver report */
	receiver_report* rr_;
	 
	
};
/** This class holds the fields of the sources. A source may only send RTP packets
   but it can be also a receiver that is active, which means it sends and receives 
   RTP packets. For the meaning of the fields of this class read RFC 3550. 
   Some fields are only for implementation cotrol and debugging
*/

class RTPSource : public TclObject {
public:
	
	RTPSource* next;
	
        /** constructor */
	RTPSource(u_int32_t srcid);
	/** returns the srcid_ of the RTPSource */
	inline u_int32_t srcid() { return (srcid_); }
	/** returns the number of packets sent np_ since last report*/
	inline int np() { return (np_); }
	/** returns the total number of packets sent snp_ since last report */
	inline int snp() { return (snp_); }
	/** returns the last highest extended number */
	inline int ehsr() { return (ehsr_); }
	/** returns the total number of bytes sent */
	inline int nbytes() { return (nbytes_); }
	/** returns the total number of packets lost */
	inline int cum_pkts_lost() { return (cum_pkts_lost_);  }
	/** returns the time of the Last Sender report */
	inline double LSR() { return (LSR_);  }
	/** returns the sender report time */
	inline double SRT() { return (SRT_);  }
	/** returns true if this source is a sender, false otherwise */
	inline bool is_sender() { return (is_sender_);  }
	/** returns the size of the RTP Packet */
	inline int ps() { return (ps_);  }
	/** retuns the sending rate in bytes/sec */
	inline double rate() { return (rate_);  }
	/** returns the difference between the RTP packet sending and receiving times */
	inline double transit() { return (transit_);  }
	/** returns the delay jitter */
	inline double jitter() { return (jitter_);  }

	/** the number of packets sent np_ since last report*/
	inline void np(int n) { np_ += n; }
	/** the total number of packets sent snp_ since last report */
	inline void snp(int n) { snp_ = n; }
	/** the last highest extended number */
	inline void ehsr(int n) { ehsr_ = n; }
	/** the total number of bytes sent */
	inline void nbytes(int n) { nbytes_+= n; }
	/** the total number of packets lost */
	inline void cum_pkts_lost(int n) { cum_pkts_lost_ += n; }
	/** the time of the Last Sender report */
	inline void LSR(double n) { LSR_= n; }
	/** the sender report time */
	inline void SRT(double n) { SRT_= n; }
	/** true if this source is a sender, false otherwise */
	inline void is_sender(bool n) { is_sender_= n; }
	/** the size of the RTP Packet */
	inline void ps(int n) { ps_= n; }
	/** the sending rate in bytes/sec */
	inline void rate(double n) { rate_= n; }
	/** the difference between the RTP packet sending and receiving times */
	inline void transit(double n) { transit_= n; }
	/** the delay jitter */
	inline void jitter(double n) { jitter_= n; }
protected:
	/** the srcid_ of the RTPSource */
	u_int32_t srcid_;
	/** the number of packets sent np_ since last report*/
	int np_;
	/** the total number of packets sent snp_ since last report */
	int snp_;
	/** the last highest extended number */
	int ehsr_;
	/** the total number of bytes sent */
	int nbytes_;
	/** the total number of packets lost */
	int cum_pkts_lost_;
	/** the time of the Last Sender report */
	double LSR_;
	/** the sender report time */
	double SRT_;
	/** true if this source is a sender, false otherwise */
	bool is_sender_;
	/** the size of the RTP Packet */
	int ps_;
	/** the size of the RTP Packet */
	double rate_;
	/** the difference between the RTP packet sending and receiving times */
	double transit_;
	/**  the delay jitter */
	double jitter_;
	
};

/** This class holds the fields of the receivers in the RTP session.
    A receiver can only receive RTP packets and it is not a source
*/
class RTPReceiver : public TclObject {
public:
	/** instance of the RTPReceiver class */
        RTPReceiver* next;
	/** constructor */
	RTPReceiver(u_int32_t srcid);
	/** the source id of the RTPReceiver */
	inline u_int32_t srcid() { return (srcid_); }
	/** returns the total number of RTP packets lost */
	inline int cum_pkts_lost() { return (cum_pkts_lost_); }
	/** returns the effective RTT time measured by the sender */
	inline double eff_rtt() { return (eff_rtt_); }
	/** returns the sendind rate in bytes/sec */
	inline double rate() { return (rate_); }
	/** adds the lost packets since the start of the session */ 
	inline void cum_pkts_lost(int n) { cum_pkts_lost_ += n; }
	/** the effective RTT as it is measured by the RTP sender */
	inline void eff_rtt(double n) { eff_rtt_ = n; }
        /** the TCP-friendly as it is reported by thr RTP receiver  */
	inline void rate(double n) { rate_ = n; }
	void receive_report(receiver_report*);
            
protected:
	/** the source id of the receiver */
	u_int32_t srcid_;
	/** the total number of packet losses since the start of the session */
	int cum_pkts_lost_;
	/** the effective RTT measured by the RTP sender */
	double eff_rtt_;
        /** the TCP-friendly as it is reported by thr RTP receiver  */
	double rate_;
};


/** The session class is the heart of the RTP protocol.
    all main functions can be found here. Any extension to the RTP 
    code should reference this class
*/
class RTPSession : public NsObject {
public:
	RTPSession();
	~RTPSession();
	/* the receive packet function */
	virtual void recv(Packet* p, Handler*);
	/* the receive control packet function */
	virtual void recv_ctrl(Packet* p);
	/* connection with TCL commands */
	int command(int argc, const char*const* argv);
	/* returns the srcid of the local source */
	inline u_int32_t srcid() { return (localsrc_->srcid()); }
	/* builds the RTCP SR and RR */
	int build_report(int bye);
	/* icrements by one each time when a RTP packet is sent */
	void localsrc_update(int);
	/* updates the trasnmission rate of the RTP sender */
	void update_rate();
	/* passes the Sender Report */
	void pass_sr(sender_report*);
	/*  increments by packet size (in bytes) each time when a RTP packet is sent */
	void localsrc_update_nbytes(int n);
	/* calculates a smooth TCP-friendly rate */
	void smooth_rate(double rate);
	RTPSource* allsrcs_;
	RTPSource* localsrc_;
	RTPReceiver* receivers_;
	/** the last packet number */
	int last_np_;
	/* we do not implement this function */
	int build_sdes();
	/* future work */
	int build_bye();
	/* access the RTP header */
	hdr_rtp* access_hdr_rtp();
	/* for debugging */
	void print_rcv();
	/** zero by default */
	int enableFlowControl_;
	/** the receiveing rate as it measured by the RTP receiver */
	double rx_recv_;
	/** the jitter delay as is it measured by the RTP receiver */
	double jitter_;


protected:
	
	RTPSource* lookup(u_int32_t);
	RTPReceiver* lookup_rcv(u_int32_t);
	/** a list to store values */
	list<double> lst;
	/* add this source */
	void enter(RTPSource*);
	/* add this receiver */
	void enter_rcv(RTPReceiver*);
	/* calculate RTT */
	void calculate_RTT();
	/* increase rate if packet losses have been not detected since last RR */
	void increase_rate(int a);
	/* measure smooth packet loss ratio */
	void measure_smooth_loss (double n);
	/* calculates TCP-friendly bandwidth share */
	void calculateR_tcp(int a);
	/* the initial session rate */
	void initial_rate(double a);
	/** the one-way trip time */
	double T_one_way_;
	/** the rtp header */
	hdr_rtp* rh_;
	/* removes the RTP receiver from the list */
	void remove_receiver(RTPReceiver*);
	/** true if the RTP sender sent packets since the last RTP SR */
	int we_sent;
	/** the RTT time */
	double RTT_;
	/** weights for the smooth loss raton calculation */
	double weight[8];
	/** the array to keep the packet loss history */
	double pkt_loss_history[8];
	/** the smooth loss ratio */
	double smooth_loss_;
	/** keeps the number of the last packets losts since the beginning of the session */
	int last_pkts_lost_;
	/** keeps the highest expected RTP sequence number */
	int last_ehsr_;
	/** the transmission rate that is measured by the receiver */
	double tx_rate_;
	/** time elapsed since the last RR (receiver report) */
	double time_elapsed_;
	/** the time when the last RR was sent */
	double last_time_report_;
	/** the value A */
	double alpha;
	/* calculates A */
	void calculate_alpha(double value);


	
};

class RTPAgent;
/**
 * the class of the RTP Timer
*/

class RTPTimer : public TimerHandler {
public: 
        RTPTimer(RTPAgent *a) : TimerHandler() { a_ = a; }
protected:
        virtual void expire(Event *e);
        RTPAgent *a_;
};
/**
 * this class generates the RTP packets and controls the transmission rate
 * through the RTP timer and the update rate function
*/
class RTPAgent : public Agent {
 public:
        RTPAgent();
        virtual void timeout(int);
        virtual void recv(Packet* p, Handler*);
        virtual int command(int argc, const char*const* argv);
        void advanceby(int delta);
        virtual void sendmsg(int nbytes, const char *flags = 0);
 protected:
        virtual void sendpkt();
	virtual void makepkt(Packet*);
        void rate_change();
        virtual void start();
        virtual void stop();
        virtual void finish();
        RTPSession* session_;
        double lastpkttime_;
        int seqno_;
        int running_;
        int random_;
        int maxpkts_;
        double interval_;
        RTPTimer rtp_timer_;
	double timestamp_;
};

#endif
