/**
 * Copyright (c) 2007 University of Patras Computer Engineering and Informatics Department Greece
 *
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
  */

#include <stdlib.h>
#include "packet.h"
#include "ip.h"
#include "rtp.h"
#include <math.h>

#define NOW Scheduler::instance().clock()

/** creates an instance of the RTPSource class
   returns the newly created object
*/
static class RTPSourceClass : public TclClass {
public:
	RTPSourceClass() : TclClass("RTPSource") {}
	TclObject* create(int argc, const char*const* argv) {
		if (argc >= 5)
			return (new RTPSource(atoi(argv[4])));

		return 0;
	}
} class_rtp_source;

/**
 *
 * creates a SessionRTP instance
 *
*/
static class RTPSessionClass : public TclClass {
public:
	RTPSessionClass() : TclClass("Session/RTP") {}
	TclObject* create(int, const char*const*) {
		return (new RTPSession());
	}
} class_rtp_session;

/**
* zero argument constructor
*/
RTPSession::RTPSession()
	: allsrcs_(0), localsrc_(0), receivers_(0),last_np_(0), T_one_way_(0),rh_(0),
	we_sent(0),last_pkts_lost_(0),last_ehsr_(-1),tx_rate_(0), time_elapsed_(0)
{
	bind("enableFlowControl_", &enableFlowControl_);
	bind("rx_recv_",&rx_recv_);
	bind("jitter_",&jitter_);
	bind("RTT_",&RTT_);
	bind("smooth_loss_",&smooth_loss_);
	alpha = 1;  //we set alpha=1 in initialization 
	last_time_report_ = NOW;

	weight[0] = 1;
	weight[1] = 1;
	weight[2] = 1;
	weight[3] = 1;
	weight[4] = 0.8;
	weight[5] = 0.6;
	weight[6] = 0.4;
	weight[7] = 0.2;
}

/**
 * destructor
*/
RTPSession::~RTPSession()
{
	while (allsrcs_ != 0) {
		RTPSource* p = allsrcs_;
		allsrcs_ = allsrcs_->next;
		delete p;
	}
	delete localsrc_;
	while (receivers_ != 0) {
                RTPReceiver* p = receivers_;
                receivers_ = receivers_->next;
                delete p;
	}
 	lst.~list<double>();
	
}

/**
 * it is called when the RTPAgent sends a RTP packet
   and updates the number of packets (np) field of the 
   originator of the RTP packet
 * @param int packet to be added
 */
void RTPSession::localsrc_update(int)
{
	localsrc_->np(1);
	localsrc_->is_sender(true);
}

/**
 *  it is called when the RTPAgent sends a RTP packet
 *  and updates the number of number of bytes (nbytes) field of the 
 *  originator of the RTP packet
 *  @param n the bytes to be added
 */
void RTPSession::localsrc_update_nbytes(int n)
{
	localsrc_->nbytes(n);
	
}

#define RTCP_HDRSIZE 8
#define RTCP_SR_SIZE 20
#define RTCP_RR_SIZE 48

/**
 *  the method for creating the SR and RR reports
 *  the "heart" of the RTCP protocol
 * @param bye if there is a bye event
 * @return  the length of this report in bytes
 */
int RTPSession::build_report(int bye)
{
	rh_ = new hdr_rtp;
	rh_->rr_ = 0;
        rh_->sr_ = 0;
        int nsrc = 0;
        int nrr = 0;
        int len = RTCP_HDRSIZE;
        we_sent = 0;
	double fraction = 0.0;
	time_elapsed_ = NOW - last_time_report_;
	last_time_report_ = NOW;
        if (localsrc_->np() != last_np_) {
                last_np_ = localsrc_->np();
                we_sent = 1;
                len += RTCP_SR_SIZE;
		//add sender report
                sender_report* sr;
                //fill the report
                sr = new sender_report;
                sr->sender_srcid()= localsrc_->srcid();
                sr->pkts_sent() = localsrc_->np();
                sr->octets_sent() = localsrc_->nbytes();
                sr->rcvrs_ = receivers_;
	        //store the report
                rh_->sr_ = sr;
        }
        for (RTPSource* sp = allsrcs_; sp != 0; sp = sp->next) {
                ++nsrc;
                int received = sp->np() - sp->snp();

                if (received == 0) {

                        continue;
                }
		if(localsrc_->srcid() != sp->srcid()) {
                int bytes = received * (sp->ps());
		rx_recv_ = ( 8 * (double)bytes)/time_elapsed_;
                sp->snp(sp->np());
                len += RTCP_RR_SIZE;
		// calculate loss fraction since previous report
		int expected_interval  = sp->ehsr() - last_ehsr_;
		last_ehsr_ = sp->ehsr();
		int lost_interval = expected_interval -  received;


		if (lost_interval <= 0 || expected_interval == 0 ) {
			 fraction = 0;

		} else fraction = ((double)lost_interval  / (double)expected_interval);
                calculate_RTT();
		// Update the R_tcp according to fraction value
		if( sp->np() >= 1) {// I have already received RTP packets from the source
			if(fraction == 0) {
				increase_rate(sp->ps());

			} 	else /*if(fraction != 0)*/ {
					measure_smooth_loss(fraction);
					calculateR_tcp(sp->ps());
				}
		}
		//add receiver report
                receiver_report* rr;
                rr = new receiver_report;
                //fill the report
		rr->cum_pkts_lost() = sp->cum_pkts_lost();
		rr->LSR() = sp->LSR();
		rr->DLSR()= NOW - sp->SRT();
		rr->R_tcp() = tx_rate_;
		rr->jitter() = sp->jitter();
		if(bye) {
		printf ("TIME for BYE %f \n", NOW);
		rr->bye()= 1;
		} else rr->bye() = 0;
		rh_->rr_ = rr;

                if (++nrr >= 31)
                        break;
		} // end if

        }// end for loop


        if (bye) {
                len += build_bye();
		}
        else
                len += build_sdes();

        Tcl::instance().evalf("%s adapt-timer %d %d %d", name(),
                              nsrc, nrr, we_sent);
        Tcl::instance().evalf("%s sample-size %d", name(), len);

        return (len);
}

/** it is called upon a bye event
   returns 8 bytes length
*/
int RTPSession::build_bye()
{
	
	return (8);
}
/** not implemented yet. It only returns 
   20 bytes that are to be added to the RTCP report
*/
int RTPSession::build_sdes()
{

	/* XXX We'll get to this later... */
	return (20);
}
/**
 *  this method accesses the RTP packets
 *  it adds a newly heard source, which is the originator of the RTP packet,
 *  calculates the one way trip time and calls the calculate_RTT(T_one_way_ , eff_rtt)
 *  method for the RTT calculation.
 *  It measures also the delay jitter time of the RTP packets.
 *  Use this method to update any wished field of the Source class instances
 * @param p the RTP packet that just received by the RTP Agent
 */
void RTPSession::recv(Packet* p, Handler*)
{
//	printf("time %f localsrc_ %d \n", Scheduler::instance().clock(),localsrc_->srcid());
//	printf("time %f localsrc_ %d \n", NOW,localsrc_->srcid());
	hdr_cmn* mh = hdr_cmn::access(p);
	hdr_rtp* rh = hdr_rtp::access(p);
    
    	RTPSource* s = lookup(rh->srcid());

    	if (s == 0) {
                if(rh->srcid()!=localsrc_->srcid()){
                        Tcl& tcl = Tcl::instance();
                        tcl.evalf("%s new-source %d", name(),rh->srcid());
                        s = (RTPSource*)TclObject::lookup(tcl.result());
		}
        }
	
	if(rh->srcid()!=localsrc_->srcid())
	{
 		// calculate one-way trip time

		T_one_way_ = NOW - rh->timestamp();
		// count packet losses
		int pkts_lost = 0;
		int difference =  rh->seqno() - s->ehsr();  
		if (difference <= 1) { // to deal with the first RTP packet in which s->ehrs() = -1
			pkts_lost = 0;
			
		}
		
		if (difference > 1) {
			pkts_lost = difference -1;
			
		}
		// calculate delay jitter, see RFC 3550 Appendix A8
		double transit = NOW - rh->timestamp();
      		double d = transit - s->transit();
      		s->transit(transit);
      		if (d < 0) d = -d;
		s->jitter( s->jitter() + (1./16.) * (d - s->jitter()));
		jitter_= s->jitter();
		s->np(1); // counts the packets that I have received from sender s
		s->cum_pkts_lost(pkts_lost);// counts the packets that I have lost from sender s
    		s->ehsr(rh->seqno());// the first ehsr = -1
    		s->nbytes(mh->size()); // counts the number of bytes that I have lost from sender s
		s->ps(mh->size());// the packet size in bytes
		
	}

        Packet::free(p);
}

/**
 *  we calculate alpha based on the RTP sender measurements of an
 *  effective RTT
 * @param value the effective RTT that is measured by the RTP sender 
 */
void RTPSession::calculate_alpha(double value)
{
	if( value == 0) {
		alpha = 1;
	} else
	alpha = (value/T_one_way_) -1;
}


/**
 *  RTT time calculation, which is performed by the receiver
 */
void RTPSession::calculate_RTT()
{
 	double temp=(1+alpha) * T_one_way_;
	RTT_ = temp * 0.1 + 0.9 * RTT_;   //taken from RFC 3448
}

/**
 *  the estimation made by the receiver of the transmission rate when
 *  fraction loss is zero. The increase is not higher than 1 RTP packet/current rate
 *  in bytes/sec
 * @param ps the packet size of the RTP packet
 */
void RTPSession::increase_rate(int ps)
{
	tx_rate_ = tx_rate_ +  (double)ps/RTT_;

}

/**
 * measure_smooth_loss is called when the last measured by the receiver
 * loss fraction is not zero.
 * @param fraction is the fraction of lost RTP packets/expected RTP packets
 * since last RR
 */
 void RTPSession::measure_smooth_loss (double fraction)
{
	
	double I_tot0 = 0;
	double I_tot1 = 0;
	double W_tot = 0;
	for (int i=0; i<7; i++) {
 	pkt_loss_history[i+1]= pkt_loss_history[i];
 	}
 	pkt_loss_history[0] = fraction;

	for (int i=0; i<7; i++) {
	I_tot0 = I_tot0+ (pkt_loss_history[i] * weight[i]);
	W_tot = W_tot +  weight[i];
	}

	for (int i=1; i<8; i++) {
	I_tot1 = I_tot1 + (pkt_loss_history[i] * weight[i-1]);
	}
	double I_tot =0;
	if( I_tot0 > I_tot1) {
	I_tot = I_tot0;
        } else {
	I_tot = I_tot1;
	}
	double  I_mean = I_tot/W_tot;

	smooth_loss_= I_mean;
	
	

}           

/**
 *  calculates the TCP friendly bandwidth share  
 * @param ps is the size of the RTP packet
 */
void RTPSession::calculateR_tcp(int ps)
{

	double min =0;
	double fraction = 3 * sqrt((3 * smooth_loss_) / 8);
	if(fraction <  1 ){
		min = fraction;
	}	else min =1;
	// TCP friendly BW share equation
	 tx_rate_ = (double)ps / (RTT_ *  (sqrt(2 * smooth_loss_ /3)) + 4 * (RTT_ * min * smooth_loss_) * (1 + (32 * pow(smooth_loss_,2))));
	 printf("Time:%f Receiver:%d smooth_loss_:%f \n", NOW,localsrc_->srcid(),smooth_loss_);

	
}

/**
 * the function that accesses and exploits the SR and RR
 *  here is where the sender calculates the effective RTT time based on 
 *  the receiver's RR. The sender also updates its transmission rate based 
 *  on the receivers' RR.
 *  We don't exploit the SR in this implementation but a third party eg
 *  a network controller can exploit both SR and RR, witout in fact joining the session
 * @param p the RTCP packet
 */
void RTPSession::recv_ctrl(Packet* p)
{
	hdr_cmn* mh = hdr_cmn::access(p);
	hdr_rtp* rh = hdr_rtp::access(p);
	u_int32_t local_src = localsrc_->srcid();
	
	if(rh->srcid() != local_src) {
	
		if (rh->sr_ != 0) { // get the Sender report
		        double temp = 0.0;
			RTPSource* source = lookup(rh->sr_->sender_srcid());
			if (source != 0 && source->srcid()!= local_src) {
				source->LSR(rh->timestamp());
				source->SRT(Scheduler::instance().clock());
				for (RTPReceiver* p = rh->sr_->rcvrs_;   p != 0; p = p->next) {
 
  					if(p->srcid() == localsrc_->srcid()) {
  						temp = p->eff_rtt();
						calculate_alpha(temp);
 					}
 				}
			}
		}

		if(localsrc_->is_sender()){ // get the Receiver report
		                 
				if(rh->rr_!= 0){
					
					RTPReceiver* s = lookup_rcv(rh->srcid());
						if (s == 0 ) {
							s = new RTPReceiver(rh->srcid());
							enter_rcv(s);
							return;
							}
						 
						if (s != 0) {
							for (s = receivers_; s!= 0; s = s->next) {
		
								if(rh->srcid() == s->srcid()) {
								double eff_rtt = 0.0;
								double alpha = Scheduler::instance().clock();
									if(rh->rr_->LSR()!=0) {
									eff_rtt = alpha - rh->rr_->LSR() - rh->rr_->DLSR();
									}
 									s->eff_rtt(eff_rtt);
                                 					s->rate(rh->rr_->R_tcp());
									if(rh->rr_->bye() == 1) remove_receiver(s);
		                                                }

							}// end for loop

						} //end if
						if(enableFlowControl_ == 1) {
						update_rate();
						}
					}
				}
			}

	Tcl::instance().evalf("%s sample-size %d", name(), mh->size());
	Packet::free(p);
}


/** this method compares the reported TCP friendly bandwidth
  * by the receivers and calls the session_bw and transmit methods
  * in session-rtp.tcl class to adjust the sender's transmission rate
*/
void RTPSession::update_rate()
{
	RTPReceiver* s;
	for (s = receivers_; s!= 0; s = s->next) {
		if(s->rate() !=0) {
			lst.push_front(s->rate());
		}
	}
	lst.sort();
	double inst_tx =lst.front();
	lst.clear();
	if(inst_tx != tx_rate_) {
		if(inst_tx < (tx_rate_/2)) {
			tx_rate_ = tx_rate_/2;
		} else tx_rate_ = inst_tx;
        printf("Time:%f Sender:%d ...Rate change to:%f \n", NOW,localsrc_->srcid(),8*tx_rate_);
	Tcl::instance().evalf("%s session_bw %fb/s",name(),8*tx_rate_);
	Tcl::instance().evalf("%s transmit %fb/s",name(),8*tx_rate_);
	}
	
}

/**
 * accesses the RTP packet header
 * @return the header of the RTP packet
 */
hdr_rtp* RTPSession::access_hdr_rtp()
{
	return rh_;
}

/**
 * it checks if the source has heard
 * @param srcid the source id of this source
 * @return the source of the RTP packet if it exists,
 * zero otherwise
 */
RTPSource* RTPSession::lookup(u_int32_t srcid)
{
	RTPSource *p;
	for (p = allsrcs_; p != 0; p = p->next)
	{	
		if (p->srcid() == srcid)
			return (p);
	}
	return (0);
}

/**
 * Check if the receiver exists in the sender's list
 * @param srcid  the source id of this receiver
 * @return  the receiver if it exists,
 *  zero otherwise
 */
RTPReceiver* RTPSession::lookup_rcv(u_int32_t srcid)
{

        RTPReceiver *p;
        for (p = receivers_; p != 0; p = p->next)
        {
                if (p->srcid() == srcid)
                        return (p);
        }
        return (0);
}

/** for debugging, prints all the receivers
    that a source serves
*/
void RTPSession::print_rcv( )
{

        RTPReceiver *p;
        for (p = receivers_; p != 0; p = p->next)
        {
            printf("Receiver:%d\n", p->srcid());
        }
}

/**
 * enter a new source
 * @param s the new RTPSource
 */
void RTPSession::enter(RTPSource* s)
{
	if(s->srcid() != localsrc_->srcid() ) {
	printf("Time:%f Receiver:%d ...I add sender %d \n", NOW,localsrc_->srcid(), s->srcid());
	s->next = allsrcs_;
	allsrcs_ = s;
	}
}

/**
 *   enter a new receiver
 * @param s the new RTPReceiver
 */
void RTPSession::enter_rcv(RTPReceiver* s)
{
	printf("Time:%f Sender:%d ...I add receiver %d \n", NOW,localsrc_->srcid(), s->srcid());
        s->next = receivers_;
        receivers_ = s;
}

/**
 * remove a receiver from a sender's list
 * @param p the RTPReceiver to be removed
 */
void RTPSession::remove_receiver(RTPReceiver* p)
{

        printf("Time:%f Sender:%d ...I remove receiver %d \n", NOW,localsrc_->srcid(),p->srcid());
	RTPReceiver* s;
	RTPReceiver* q;
        if (receivers_==p)
        {
            	q = receivers_;
                receivers_=receivers_->next;
                delete q;
        }
        else
        {
        	for (s = receivers_; s!=0; s=s->next)
        	{
        		if (s->next==p)
                	{
                 		q = s->next;
                 		s->next=q->next;
                 		delete q; 
                 	}	
                }
        }

}

/**
 * it's good for the receiver to have
   a rough knowledge of the session rate
 * @param a a rough estimation of the session rate
 */
void RTPSession::initial_rate(double a)
{
	tx_rate_ =(double)a/8;
	
	
}

/**
  * the command function that handles TCL commands
 *
 */
int RTPSession::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "enter") == 0) {
			RTPSource* s = (RTPSource*)TclObject::lookup(argv[2]);
			enter(s);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "localsrc") == 0) {
			localsrc_ = (RTPSource*)TclObject::lookup(argv[2]);
			enter(localsrc_);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "rate") == 0) {
			initial_rate(atoi(argv[2]));
			return(TCL_OK);
		}
	}

	return (TclObject::command(argc, argv));
}

/**
 * Constructor
 * @param srcid the source id of the newly created RTPReceiver
 */
RTPReceiver::RTPReceiver(u_int32_t srcid)
	: next(0),cum_pkts_lost_(0),eff_rtt_(0),rate_(0)
{
	srcid_ = srcid;
}

/**
 * Constructor of the RTPSource class
 * @param srcid the source id of the newly created RTPSource
 */
RTPSource::RTPSource(u_int32_t srcid)
	: next(0),np_(0), snp_(0), ehsr_(-1), nbytes_(0),cum_pkts_lost_(0),LSR_(0),SRT_(0),is_sender_(false),ps_(0),rate_(0),transit_(0),jitter_(0)
{
	bind("srcid_", (int*)&srcid_);
	srcid_ = srcid;
}





